from typing import cast

from .auths import ConsumerAuth, auth_factory
from .components import CrudComponent, JsonType, KongError, KongResponseError
from .plugins import KongEntityWithPlugins


class Consumer(KongEntityWithPlugins):
    @property
    def username(self) -> str:
        return self.data.get("username", "")

    @property
    def acls(self) -> CrudComponent:
        return CrudComponent(self, "acls")

    @property
    def jwts(self) -> ConsumerAuth:
        return auth_factory(self, "jwt")

    @property
    def keyauths(self) -> ConsumerAuth:
        return auth_factory(self, "key-auth")

    @property
    def basicauths(self) -> ConsumerAuth:
        return auth_factory(self, "basic-auth")


class Consumers(CrudComponent):
    Entity = Consumer

    async def apply_credentials(self, auths: list[dict], consumer: Consumer) -> None:
        for auth_data in auths:
            auth = auth_factory(consumer, auth_data["type"])
            await auth.create_or_update_credentials(auth_data["config"])

    async def apply_json(self, data: JsonType, clear: bool = True) -> list:
        if not isinstance(data, list):
            data = [data]
        result = []
        for entry in data:
            if not isinstance(entry, dict):
                raise KongError("dictionary required")
            entry = entry.copy()
            groups = entry.pop("groups", [])
            auths = entry.pop("auths", [])
            udata = entry.copy()
            id_ = udata.pop("id", None)
            username = None
            if not id_:
                username = udata.pop("username", None)
                if not username:
                    raise KongError("Consumer username or id is required")
            uid = cast(str, id_ or username)
            try:
                entity = await self.get(uid)
            except KongResponseError as exc:
                if exc.status == 404:
                    entity = await self.create(**entry)
                else:
                    raise
            else:
                if entry:
                    entity = await self.update(uid, **udata)
            consumer = cast(Consumer, entity)
            acls = await consumer.acls.get_list()
            current_groups = dict(((a["group"], a) for a in acls))
            for group in groups:
                if group not in current_groups:
                    await consumer.acls.create(group=group)
                else:
                    current_groups.pop(group)

            for acl in current_groups.values():
                await consumer.acls.delete(acl["id"])

            await self.apply_credentials(auths, consumer)

            result.append(consumer.data)

        return result



from typing import cast

from .components import UUID, CrudComponent, JsonType, KongError
from .plugins import KongEntityWithPlugins
from .routes import Routes
from .utils import local_ip

REMOVE = frozenset(("absent", "remove"))
LOCAL_HOST = frozenset(("localhost", "127.0.0.1"))


class Service(KongEntityWithPlugins):
    """Object representing a Kong service"""

    @property
    def routes(self) -> Routes:
        return Routes(self)

    @property
    def host(self) -> str:
        return self.data.get("host", "")


class Services(CrudComponent):
    """Kong Services"""

    Entity = Service

    async def delete(self, id_: str | UUID) -> bool:
        srv = cast(Service, self.wrap({"id": id_}))
        await srv.routes.delete_all()
        await srv.plugins.delete_all()
        return await super().delete(id_)

    async def apply_json(self, data: JsonType, clear: bool = True) -> list:
        """Apply a JSON data objects for services"""
        if not isinstance(data, list):
            data = [data]
        result = []
        for entry in data:
            if not isinstance(entry, dict):
                raise KongError("dictionary required")
            entry = entry.copy()
            ensure = entry.pop("ensure", None)
            name = entry.pop("name", None)
            id_ = entry.pop("id", None)
            id_or_name = name or id_
            routes = entry.pop("routes", [])
            plugins = entry.pop("plugins", [])
            host = entry.pop("host", None)
            if host in LOCAL_HOST:
                host = local_ip()
            if ensure in REMOVE:
                if not id_or_name:
                    raise KongError(
                        "Service name or id is required to remove previous services"
                    )
                if await self.has(id_or_name):
                    await self.delete(id_or_name)
                continue
            entry.update(host=host)
            if id_or_name and await self.has(id_or_name):
                if id_ and name:
                    entry.update(name=name)
                entity = await self.update(id_or_name, **entry)
            else:
                if name:
                    entry.update(name=name)
                entity = await self.create(**entry)
            srv = cast(Service, entity)
            srv.data["routes"] = await srv.routes.apply_json(routes)
            srv.data["plugins"] = await srv.plugins.apply_json(plugins)
            result.append(srv)
        return result



import asyncio
import click
import json
import yaml as _yaml
from typing import Any, cast

from . import __version__
from .client import Kong, KongError
from .consumers import Consumer
from .utils import local_ip


@click.command()
@click.option("--version", is_flag=True, default=False, help="Display version and exit")
@click.option("--ip", is_flag=True, default=False, help="Show local IP address")
@click.option(
    "--key-auth", help="Create or display an authentication key for a consumer"
)
@click.option("--yaml", type=click.File("r"), help="Yaml configuration to upload")
@click.option(
    "--clear", default=False, is_flag=True, help="Clear objects not in configuration"
)
@click.pass_context
def kong(
    ctx: click.Context,
    version: bool,
    ip: bool,
    key_auth: str,
    yaml: click.File | None,
    clear: bool,
) -> None:
    if version:
        click.echo(__version__)
    elif ip:
        click.echo(local_ip())
    elif key_auth:
        _run(_auth_key(key_auth))
    elif yaml:
        _run(_yml(yaml, clear))
    else:
        click.echo(ctx.get_help())


def _run(coro: Any) -> None:
    asyncio.get_event_loop().run_until_complete(coro)


async def _yml(yaml: Any, clear: bool) -> None:
    async with Kong() as cli:
        try:
            result = await cli.apply_json(_yaml.safe_load(yaml), clear=clear)
            click.echo(json.dumps(result, indent=4))
        except KongError as exc:
            raise click.ClickException(str(exc))


async def _auth_key(consumer: str) -> None:
    async with Kong() as cli:
        try:
            c = cast(Consumer, await cli.consumers.get(consumer))
            keys = await c.keyauths.get_list()
            if keys:
                key = keys[0]
            else:
                key = await c.keyauths.create()
            click.echo(json.dumps(key.data, indent=4))
        except KongError as exc:
            raise click.ClickException(str(exc))



from .components import CrudComponent, JsonType


class Snis(CrudComponent):
    """Kong SNI API component"""

    async def apply_json(self, data: JsonType, clear: bool = True) -> list:
        """Apply a JSON data objects for snis - never clear them"""
        if not isinstance(data, list):
            data = [data]
        result = []
        for entry in data:
            entry = entry.copy()
            name = entry.pop("name")
            if await self.has(name):
                sni = await self.update(name, **entry)
            else:
                sni = await self.create(name=name, **entry)
            result.append(sni.data)
        return result



from __future__ import annotations

import os
import sys
from typing import Any, Callable, Dict, Optional

from aiohttp import ClientResponse, ClientSession

from . import __version__
from .certificates import Certificates
from .components import CrudComponent, KongError, KongResponseError
from .consumers import Consumers
from .plugins import Plugins
from .routes import Routes
from .services import Services
from .snis import Snis

__all__ = ["Kong", "KongError", "KongResponseError"]

DEFAULT_USER_AGENT = (
    f"Python/{'.'.join(map(str, sys.version_info[:2]))} aio-kong/{__version__}"
)


class Kong:
    """Kong client"""

    url: str = os.getenv(
        "KONG_ADMIN_URL", os.getenv("KONG_URL", "http://127.0.0.1:8001")
    )
    content_type: str = "application/json, text/*; q=0.5"

    def __init__(
        self,
        url: str = "",
        session: ClientSession | None = None,
        request_kwargs: dict | None = None,
        user_agent: str = DEFAULT_USER_AGENT,
    ) -> None:
        self.url = url or self.url
        self.session = session
        self.user_agent = user_agent
        self.request_kwargs = request_kwargs or {}
        self.services = Services(self)
        self.routes = Routes(self)
        self.plugins = Plugins(self)
        self.consumers = Consumers(self)
        self.certificates = Certificates(self)
        self.acls = CrudComponent(self, "acls")
        self.snis = Snis(self)

    def __repr__(self) -> str:
        return self.url

    __str__ = __repr__

    @property
    def cli(self) -> Kong:
        return self

    async def close(self) -> None:
        if self.session:
            await self.session.close()

    async def __aenter__(self) -> Kong:
        return self

    async def __aexit__(self, exc_type: type, exc_val: Any, exc_tb: Any) -> None:
        await self.close()

    async def execute(
        self,
        url: str,
        method: str = "",
        headers: dict[str, str] | None = None,
        callback: Optional[Callable[[ClientResponse], Any]] = None,
        wrap: Optional[Callable[[Any], Any]] = None,
        **kw: Any,
    ) -> Any:
        if not self.session:
            self.session = ClientSession()
        method = method or "GET"
        headers_ = self.default_headers()
        headers_.update(headers or ())
        kw.update(self.request_kwargs)
        response = await self.session.request(method, url, headers=headers_, **kw)
        if callback:
            return await callback(response)
        if response.status == 204:
            return True
        if response.status >= 400:
            try:
                data = await response.json()
            except Exception:
                data = await response.text()
            raise KongResponseError(response, data)
        response.raise_for_status()
        data = await response.json()
        return wrap(data) if wrap else data

    async def apply_json(self, config: dict, clear: bool = True) -> dict:
        if not isinstance(config, dict):
            raise KongError("Expected a dict got %s" % type(config).__name__)
        result = {}
        for name, data in config.items():
            if not isinstance(data, list):
                data = [data]
            o = getattr(self, name)
            if not o:
                raise KongError("Kong object %s not available" % name)
            result[name] = await o.apply_json(data, clear=clear)
        return result

    async def delete_all(self) -> None:
        await self.services.delete_all()
        await self.consumers.delete_all()
        await self.plugins.delete_all()
        await self.certificates.delete_all()

    def default_headers(self) -> Dict[str, str]:
        return {"user-agent": self.user_agent, "accept": self.content_type}



from .components import CrudComponent, KongEntity
from .snis import Snis


class Certificate(KongEntity):
    @property
    def snis(self) -> Snis:
        return Snis(self)


class Certificates(CrudComponent):
    """Kong TLS certificate component"""

    Entity = Certificate



from __future__ import annotations

import json
from aiohttp import ClientResponse
from typing import TYPE_CHECKING, Any, AsyncIterator, Iterator, Mapping

from .utils import UUID, as_dict, as_params, uid

if TYPE_CHECKING:
    from .client import Kong


JsonType = dict | list


class KongError(Exception):
    pass


class KongResponseError(KongError):
    def __init__(self, response: ClientResponse, message: str = "") -> None:
        self.response = response
        self.message = as_dict(message, "message")
        self.message["request_url"] = str(response.url)
        self.message["request_method"] = response.method
        self.message["response_status"] = response.status

    @property
    def status(self) -> int:
        return self.response.status

    def __str__(self) -> str:
        return json.dumps(self.message, indent=4)


class KongEntity(Mapping[str, Any]):
    """A Kong entity is either a

    - Service
    - Route
    - Plugin
    - Consumer
    - Certificate
    - SNI
    """

    def __init__(self, root: Kong | CrudComponent, data: dict[str, Any]) -> None:
        self.root = root
        self.data = data

    def __repr__(self) -> str:
        return repr(self.data)

    def __str__(self) -> str:
        return self.__repr__()

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterator[str]:
        return iter(self.data)

    def __getitem__(self, item: Any) -> Any:
        return self.data[item]

    def __contains__(self, item: Any) -> bool:
        return item in self.data

    @property
    def cli(self) -> Kong:
        return self.root.cli

    @property
    def id(self) -> str:
        return self.data["id"]

    @property
    def name(self) -> str:
        return self.data.get("name") or ""

    @property
    def url(self) -> str:
        return "%s/%s" % (self.root.url, self.id)

    def get(self, item: Any, default: Any = None) -> Any:
        return self.data.get(item, default)

    async def execute(self, url: str, method: str = "", **params: Any) -> Any:
        return await self.root.execute(url, method, **params)


class CrudComponent:
    Entity = KongEntity

    def __init__(self, root: Kong | KongEntity, name: str = "") -> None:
        self.root = root
        self.name = name or self.__class__.__name__.lower()

    def __repr__(self) -> str:
        return self.url

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def cli(self) -> Kong:
        return self.root.cli

    @property
    def url(self) -> str:
        return f"{self.cli.url}/{self.name}"

    @property
    def is_entity(self) -> bool:
        return isinstance(self.root, KongEntity)

    async def execute(self, url: str, method: str = "", **kwargs: Any) -> Any:
        return await self.root.execute(url, method, **kwargs)

    async def apply_json(self, data: JsonType, clear: bool = True) -> list:
        raise NotImplementedError

    async def paginate(self, **params: Any) -> AsyncIterator[KongEntity]:
        url = self.list_create_url()
        next_ = url
        exec_params = as_params(**params)
        while next_:
            if not next_.startswith(url):
                next_ = f'{url}?{next_.split("?")[1]}'
            data = await self.execute(next_, params=exec_params)
            next_ = data.get("next")
            for d in data["data"]:
                yield self.wrap(d)

    async def get_list(self, **params: Any) -> list[KongEntity]:
        url = self.list_create_url()
        return await self.execute(url, params=as_params(**params), wrap=self.wrap_list)

    async def get_full_list(self, **params: Any) -> list[KongEntity]:
        return [d async for d in self.paginate(**params)]

    async def get(self, id_: str | UUID) -> KongEntity:
        url = f"{self.url}/{uid(id_)}"
        return await self.execute(url, wrap=self.wrap)

    async def has(self, id_: str | UUID) -> bool:
        url = f"{self.url}/{uid(id_)}"
        return await self.execute(url, "get", callback=self.head)

    async def create(self, **params: Any) -> KongEntity:
        url = self.list_create_url()
        return await self.execute(url, "post", json=params, wrap=self.wrap)

    async def update(self, id_: str | UUID, **params: Any) -> KongEntity:
        url = f"{self.url}/{uid(id_)}"
        return await self.execute(url, "patch", json=params, wrap=self.wrap)

    async def delete(self, id_: str | UUID) -> bool:
        url = f"{self.url}/{uid(id_)}"
        return await self.execute(url, "delete")

    async def delete_all(self) -> int:
        n = 0
        async for entity in self.paginate():
            await self.delete(entity.id)
            n += 1
        return n

    async def head(self, response: ClientResponse) -> bool:
        if response.status == 404:
            return False
        elif response.status == 200:
            return True
        else:  # pragma: no cover
            raise KongResponseError(response)

    def wrap(self, data: dict) -> KongEntity:
        return self.Entity(self, data)

    def wrap_list(self, data: dict) -> list[KongEntity]:
        return [self.wrap(d) for d in data["data"]]

    def list_create_url(self) -> str:
        if self.is_entity:
            return f"{self.root.url}/{self.name}"
        else:
            return self.url



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .components import CrudComponent, KongEntity

if TYPE_CHECKING:
    from .consumers import Consumer


def auth_factory(consumer: Consumer, auth_type: str) -> ConsumerAuth:
    known_types = {"basic-auth": BasicAuth, "key-auth": KeyAuth}
    constructor = known_types.get(auth_type, ConsumerAuth)
    return constructor(consumer, auth_type)


class ConsumerAuth(CrudComponent):
    unique_field: str = ""

    @property
    def url(self) -> str:
        return f"{self.root.url}/{self.name}"

    async def get_existing_id(self, creds_config: dict) -> str | None:
        if not self.unique_field:
            raise NotImplementedError(
                "Existence check not implemented for this type of\
                 authentication"
            )
        cur_unique = creds_config[self.unique_field]
        try:
            return next(
                cred
                for cred in await self.get_list()
                if cred[self.unique_field] == cur_unique
            )["id"]
        except StopIteration:
            return None

    async def create_or_update_credentials(self, creds_config: dict) -> KongEntity:
        if existing_id := await self.get_existing_id(creds_config):
            return await self.update_credentials(existing_id, data=creds_config)
        else:
            return await self.create_credentials(data=creds_config)

    async def update_credentials(self, id_: str, **kw: Any) -> KongEntity:
        url = f"{self.url}/{id_}"

        return await self.cli.execute(
            url,
            "patch",
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            wrap=self.wrap,
            **kw,
        )

    async def create_credentials(self, **kw: Any) -> KongEntity:
        return await self.cli.execute(
            self.url,
            "post",
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            wrap=self.wrap,
            **kw,
        )

    async def get_or_create(self) -> KongEntity:
        secrets = await self.get_list(limit=1)
        return secrets[0] if secrets else await self.create()


class BasicAuth(ConsumerAuth):
    unique_field = "username"


class KeyAuth(ConsumerAuth):
    unique_field = "key"
